# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import torch
import ocnn

from typing import Optional
from ocnn.octree import Octree, key2xyz, xyz2key


class Graph:

  def __init__(self, **kwargs):
    super().__init__()
    # !!! Important !!!
    # Each key has 64 bits. The first 48 bits contains the shuffled key, the
    # 49th to 58th bits (10 bits) contains the batch index, and the 59th to 63th
    # bits (5 bits) contains the node depth. Therefore, the batch size should be
    # smaller than 2^10, and the octree depth should be smaller than 2^5.
    self.key_shift = 58
    self.depth_min = None
    self.depth_max = None
    self.nnum = None         # the total node number

    self.batch_id = None
    self.node_depth = None
    self.child = None        # the octree node has children or not
    self.key = None          # the bits from the 58th is the node depth
    self.octree_mask = None  # used to pad zeros for non-leaf nodes

    self.edge_idx = None     # the edge index in (i, j)
    self.edge_dir = None     # the dirction of the edge

  @property
  def node_type(self):
    return self.node_depth - self.depth_min

  @property
  def edge_type(self):
    return self.edge_dir


class OctreeD(Octree):

  def __init__(self, octree: Octree, max_depth: Optional[int] = None, **kwargs):
    super().__init__(octree.depth, octree.full_depth)
    self.__dict__.update(octree.__dict__)

    # node numbers for the octree
    self._set_node_num()

    # for the construction of the dual octree
    self.graphs = [Graph() for _ in range(self.depth + 1)]

    # build the dual octree
    self._build_lookup_tables()
    self.build_dual_graph(max_depth)

  def _set_node_num(self):
    self.ncum = ocnn.utils.cumsum(self.nnum, dim=0, exclusive=True)
    self.lnum = self.nnum - self.nnum_nempty  # leaf node numbers
    self.lnum[self.depth] = self.nnum[self.depth]

  def _build_lookup_tables(self):
    self.ngh = torch.tensor(
        [[0, 0, 1], [0, 0, -1],       # up, down
         [0, 1, 0], [0, -1, 0],       # right, left
         [1, 0, 0], [-1, 0, 0]],      # front, back
        dtype=torch.int16, device=self.device)
    self.dir_table = torch.tensor(
        [[1, 3, 5, 7], [0, 2, 4, 6],   # up, down
         [2, 3, 6, 7], [0, 1, 4, 5],   # right, left
         [4, 5, 6, 7], [0, 1, 2, 3]],  # front, back
        dtype=torch.int64, device=self.device)
    self.dir_type = torch.tensor(
        [0, 1, 2, 3, 4, 5],
        dtype=torch.int64, device=self.device)
    self.dir_remap = torch.tensor(
        [1, 0, 3, 2, 5, 4],
        dtype=torch.int64, device=self.device)
    self.interal_row = torch.tensor(
        [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3,
         4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7],
        dtype=torch.int64, device=self.device)
    self.interal_col = torch.tensor(
        [1, 2, 4, 0, 3, 5, 0, 3, 6, 1, 2, 7,
         0, 5, 6, 1, 4, 7, 2, 4, 7, 3, 5, 6],
        dtype=torch.int64, device=self.device)
    self.interal_dir = torch.tensor(
        [0, 2, 4, 1, 2, 4, 3, 0, 4, 3, 1, 4,
         5, 0, 2, 5, 1, 2, 5, 3, 0, 5, 3, 1],
        dtype=torch.int64, device=self.device)

  def batch_id(self, depth: int, nempty: bool = False):
    r""" Override the method in the Octree class. """
    return self.graphs[depth].batch_id

  def arange(self, num: int):
    return torch.arange(num, dtype=torch.int64, device=self.device)

  def empty_tensor(self):
    return torch.tensor([], dtype=torch.int64, device=self.device)

  def build_dual_graph(self, max_depth: Optional[int] = None):
    depth = max_depth or self.depth
    self.dense_graph(self.full_depth)
    for d in range(self.full_depth + 1, depth + 1):
      self.sparse_graph(d)

  def octree_split(self, split: torch.Tensor, depth: int):
    #  update self.children[depth], self.nnum_nempty[depth]
    super().octree_split(split, depth)

    # update the properties of the dual octree
    self._set_node_num()
    self.graphs[depth].child[-self.nnum[depth]:] = self.children[depth]

  def octree_grow(self, depth: int, update_neigh: bool = True):
    super().octree_grow(depth, update_neigh)
    if depth > self.full_depth:
      self.sparse_graph(depth)
    else:
      self.dense_graph(depth)

  def dense_graph(self, depth: int = 3):
    K = 6  # each node has at most K neighboring node
    bnd = 2 ** depth
    num = bnd ** 3

    ki = self.arange(num)
    x, y, z, _ = key2xyz(ki, depth)
    xi = torch.stack([x, y, z], dim=1)  # [N, 3]
    xj = xi.unsqueeze(1) + self.ngh     # [N, K, 3]
    xj = xj.view(-1, 3)                 # [N*K, 3]

    # for full octree, the octree key is the index
    row = ki.unsqueeze(1).repeat(1, K).view(-1)               # [N*K]
    col = xyz2key(xj[:, 0], xj[:, 1], xj[:, 2], depth=depth)  # [N*K]
    edge_dir = self.dir_type.repeat(num)                      # [K] -> [N*K]

    # remove invalid edges
    valid = torch.logical_and(xj > -1, xj < bnd)  # out-of-bound
    valid = torch.all(valid, dim=-1).view(-1)
    row, col, edge_dir = row[valid], col[valid], edge_dir[valid]

    # deal with batches
    dis = self.arange(self.batch_size)
    dis = dis.unsqueeze(1) * num
    row = row.unsqueeze(0) + dis
    col = col.unsqueeze(0) + dis
    edge_dir = edge_dir.repeat(self.batch_size)

    # graph edges
    graph = self.graphs[depth]
    graph.edge_dir = edge_dir.view(-1)
    graph.edge_idx = torch.stack([row.view(-1), col.view(-1)])

    # graph nodes
    node_key = self.keys[depth]
    graph.batch_id = node_key >> 48
    graph.node_depth = torch.ones_like(node_key) * depth
    graph.key = torch.bitwise_or(node_key, graph.node_depth << graph.key_shift)
    graph.child = self.children[depth]
    ones = torch.ones(self.nnum[depth], dtype=bool, device=self.device)
    graph.octree_mask = ones

    # depth and node numbers
    graph.depth_max = depth
    graph.depth_min = depth
    graph.nnum = node_key.numel()

  def sparse_graph(self, depth: int):
    # get the previous graph
    graph = self.graphs[depth - 1]
    edge_idx = graph.edge_idx
    edge_dir = graph.edge_dir
    leaf_mask = graph.child < 0
    idx_bias = graph.nnum

    # mark invalid nodes of layer (depth-1)
    row, col = edge_idx[0], edge_idx[1]
    valid_row = graph.child[row] < 0
    valid_col = graph.child[col] < 0
    invalid_row = torch.logical_not(valid_row)
    invalid_col = torch.logical_not(valid_col)
    valid_edges = torch.logical_and(valid_row, valid_col)
    invalid_row_vtx = torch.logical_and(invalid_row, valid_col)
    invalid_both_vtx = torch.logical_and(invalid_row, invalid_col)

    # deal with edges with invalid row vtx only
    vi = row[invalid_row_vtx]
    vj = col[invalid_row_vtx]
    di = edge_dir[invalid_row_vtx]
    row_o1 = graph.child[vi].unsqueeze(1) * 8 + self.dir_table[di, :]
    row_o1 = row_o1.view(-1) + idx_bias  # NOTE: add graph.nnum
    col_o1 = vj.unsqueeze(1).repeat(1, 4).view(-1)
    dir_o1 = di.unsqueeze(1).repeat(1, 4).view(-1)

    # deal with edges with 2 invalid nodes
    if invalid_both_vtx.any():
      vi = row[invalid_both_vtx]
      vj = col[invalid_both_vtx]
      di = edge_dir[invalid_both_vtx]
      dj = self.dir_remap[di]
      row_o2 = graph.child[vi].unsqueeze(1) * 8 + self.dir_table[di, :]
      row_o2 = row_o2.view(-1) + idx_bias  # NOTE: add graph.nnum
      col_o2 = graph.child[vj].unsqueeze(1) * 8 + self.dir_table[dj, :]
      col_o2 = col_o2.view(-1) + idx_bias  # NOTE: add graph.nnum
      dir_o2 = di.unsqueeze(1).repeat(1, 4).view(-1)
    else:
      row_o2 = self.empty_tensor()
      col_o2 = self.empty_tensor()
      dir_o2 = self.empty_tensor()

    # add internal edges connecting sibling nodes.
    row_i, col_i, dir_i = self._internal_edges(depth, idx_bias)

    # update graph edges
    graph_d = self.graphs[depth]
    graph_d.edge_idx = torch.stack([
        torch.cat([row[valid_edges], row_i, row_o1, col_o1, row_o2]),
        torch.cat([col[valid_edges], col_i, col_o1, row_o1, col_o2])])
    graph_d.edge_dir = torch.cat([
        edge_dir[valid_edges], dir_i, dir_o1, self.dir_remap[dir_o1], dir_o2])

    # update graph nodes
    node_key = self.keys[depth]
    batch_id = node_key >> 48
    node_depth = torch.ones_like(node_key) * depth
    node_key = torch.bitwise_or(node_key, node_depth << 58)
    leaf_mask = graph.child < 0
    graph_d.batch_id = torch.cat([graph.batch_id[leaf_mask], batch_id])
    graph_d.node_depth = torch.cat([graph.node_depth[leaf_mask], node_depth])
    graph_d.key = torch.cat([graph.key[leaf_mask], node_key])
    graph_d.child = torch.cat([graph.child[leaf_mask], self.children[depth]])
    octree_mask = torch.cat(self.children[graph.depth_min:depth]) < 0
    ones = torch.ones(self.nnum[depth], dtype=bool, device=self.device)
    graph_d.octree_mask = torch.cat([octree_mask, ones])

    # remap the edge index since some nodes are removed
    mask = torch.cat([leaf_mask, ones], dim=0)
    remapper = torch.cumsum(mask.long(), dim=0) - 1
    graph_d.edge_idx = remapper[graph_d.edge_idx]

    # depth and node numbers
    graph_d.depth_max = depth
    graph_d.depth_min = graph.depth_min
    graph_d.nnum = graph_d.key.numel()

  def _internal_edges(self, depth: int, idx_bias: int):
    num = int(self.nnum[depth] / 8)
    dis = self.arange(num).unsqueeze(1) * 8 + idx_bias
    row = self.interal_row.unsqueeze(0) + dis
    col = self.interal_col.unsqueeze(0) + dis
    edge_dir = self.interal_dir.repeat(num)
    return row.view(-1), col.view(-1), edge_dir.view(-1)

  def pad_zeros(self, data: torch.Tensor, depth: int):
    r''' Pad zeros to the input data so that the length of the feature tensor is
    the same as the number of leaf nodes before the layer depth.
    '''

    leaf_num = torch.sum(self.lnum[self.full_depth:depth])
    zeros = data.new_zeros([leaf_num, data.shape[1]])
    data = torch.cat([zeros, data], dim=0)
    return data

  def get_input_feature(self, all_leaf_nodes: bool = True, feature: str = 'L'):
    r''' `all_leaf_nodes` is used to determine whether to concat zero features
    with the initial features in layer depth. If `all_leaf_nodes=False`, the
    behavior is the same as the original `Octree.get_input_feature`; otherwise,
    the zero features will be concatenated so that the length of the feature
    tensor is the same as the number of **all** leaf nodes, which is expected
    when using ognn's convolution layers.
    '''

    # the initial feature of leaf nodes in the layer depth
    data = super().get_input_feature(feature)

    # concat zero features with the initial features in layer depth
    if all_leaf_nodes:
      data = self.pad_zeros(data, self.depth)
    return data



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import os
import torch
import numpy as np

from thsolver import Dataset
from .utils import collate_func
from .shapenet import TransformShape


class TransformScene(TransformShape):

  def __init__(self, flags):
    super().__init__(flags)

    self.point_sample_num = 10000
    self.occu_sample_num = 4096
    self.surface_sample_num = 2048
    self.sample_surf_points = flags.sample_surf_points
    self.points_scale = 0.6  # the points are actually in [-0.55, 0.55]
    self.noise_std = 0.005
    self.pos_weight = 10

  def sample_occu(self, sample):
    points, occus = sample['points'], sample['occupancies']
    points = points / self.points_scale
    points = points + 1e-6 * np.random.randn(*points.shape)  # ref ConvoNet
    occus = np.unpackbits(occus)[:points.shape[0]]

    rand_idx = np.random.choice(points.shape[0], size=self.occu_sample_num)
    points = torch.from_numpy(points[rand_idx]).float()
    occus = torch.from_numpy(occus[rand_idx]).float()
    # 1 - outside shapes; 1 - inside shapes
    occus = 1 - occus  # to be consistent with ShapeNet

    # The number of points inside shapes is roughly 1.2% of points outside
    # shapes, we set weights to 10 for points inside shapes.
    weight = torch.ones_like(occus)
    weight[occus < 0.5] = self.pos_weight

    # The points are not on the surfaces, the gradients are 0
    grad = torch.zeros_like(points)
    return {'pos': points, 'occu': occus, 'weight': weight, 'grad': grad}

  def sample_surface(self, sample):
    # get the input TODO: use normals
    points, normals = sample['points'], sample['normals']

    # sample points
    rand_idx = np.random.choice(points.shape[0], size=self.surface_sample_num)
    pos = torch.from_numpy(points[rand_idx])
    grad = torch.from_numpy(normals[rand_idx])
    pos = pos / self.points_scale  # scale to [-1.0, 1.0]
    occus = torch.ones(self.surface_sample_num) * 0.5
    weight = torch.ones(self.surface_sample_num) * 2.0  # TODO: tune this scale

    return {'pos': pos, 'occu': occus, 'weight': weight, 'grad': grad}

  def __call__(self, sample, idx):
    output = self.process_points_cloud(sample['point_cloud'])

    # sample ground truth sdfs
    if self.flags.load_occu:
      occus = self.sample_occu(sample['occus'])

      if self.sample_surf_points:
        surface_occus = self.sample_surface(sample['point_cloud'])
        for key in occus.keys():
          occus[key] = torch.cat([occus[key], surface_occus[key]], dim=0)

      output.update(occus)

    return output


class ReadFile:
  def __init__(self, load_occu=False):
    self.load_occu = load_occu
    self.num_files = 10

  def __call__(self, filename):
    num = np.random.randint(self.num_files)
    filename_pc = os.path.join(
        filename, 'pointcloud/pointcloud_%02d.npz' % num)
    raw = np.load(filename_pc)
    point_cloud = {'points': raw['points'], 'normals': raw['normals']}
    output = {'point_cloud': point_cloud}

    if self.load_occu:
      num = np.random.randint(self.num_files)
      filename_occu = os.path.join(
          filename, 'points_iou/points_iou_%02d.npz' % num)
      raw = np.load(filename_occu)
      occus = {'points': raw['points'], 'occupancies': raw['occupancies']}
      output['occus'] = occus

    return output


def get_synthetic_room_dataset(flags):
  transform = TransformScene(flags)
  read_file = ReadFile(flags.load_occu)
  dataset = Dataset(flags.location, flags.filelist, transform,
                    read_file=read_file, in_memory=flags.in_memory)
  return dataset, collate_func



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import ocnn
import torch
import torch.nn

from ognn import mpu, nn
from ognn.octreed import OctreeD


class GraphOUNet(torch.nn.Module):

  def __init__(self, in_channels: int, resblk_type: str = 'basic',
               feature: str = 'L', norm_type: str = 'batch_norm',
               act_type: str = 'relu', **kwargs):
    super().__init__()
    self.bottleneck = 4
    self.n_edge_type = 7
    self.head_channel = 64
    self.in_channels = in_channels
    self.resblk_type = resblk_type
    self.feature = feature
    self.norm_type = norm_type
    self.act_type = act_type
    self.config_network()

    self.graph_pad = nn.GraphPad()
    self.encoder_stages = len(self.encoder_blk_nums)
    self.decoder_stages = len(self.decoder_blk_nums)

    # encoder
    n_node_types = [self.n_node_type - i for i in range(self.encoder_stages)]
    self.conv1 = nn.GraphConvNormAct(
        in_channels, self.encoder_channels[0], self.n_edge_type,
        n_node_types[0], self.norm_type, self.act_type)
    self.encoder = torch.nn.ModuleList([nn.GraphResBlocks(
        self.encoder_channels[i], self.encoder_channels[i],
        self.n_edge_type, n_node_types[i], self.norm_type,
        self.act_type, self.bottleneck, self.encoder_blk_nums[i],
        self.resblk_type) for i in range(self.encoder_stages)])
    self.downsample = torch.nn.ModuleList([nn.GraphDownsample(
        self.encoder_channels[i], self.encoder_channels[i+1],
        self.norm_type, self.act_type) for i in range(self.encoder_stages - 1)])

    # decoder
    n_node_type = self.n_node_type - self.encoder_stages + 1
    n_node_types = [n_node_type + i for i in range(self.decoder_stages)]
    self.upsample = torch.nn.ModuleList([nn.GraphUpsample(
        self.decoder_channels[i - 1], self.decoder_channels[i],
        self.norm_type, self.act_type) for i in range(1, self.decoder_stages)])
    self.decoder = torch.nn.ModuleList([nn.GraphResBlocks(
        self.decoder_channels[i], self.decoder_channels[i],
        self.n_edge_type, n_node_types[i], self.norm_type,
        self.act_type, self.bottleneck, self.decoder_blk_nums[i],
        self.resblk_type) for i in range(1, self.decoder_stages)])

    # header
    mid_channels = 32
    self.predict = torch.nn.ModuleList([nn.Prediction(
        self.decoder_channels[i], mid_channels, 2, self.norm_type,
        self.act_type) for i in range(self.decoder_stages)])
    self.regress = torch.nn.ModuleList([nn.Prediction(
        self.decoder_channels[i], mid_channels, 4, self.norm_type,
        self.act_type) for i in range(self.decoder_stages)])

  def config_network(self):
    self.n_node_type = 5
    self.encoder_blk_nums = [3, 3, 3, 3]
    self.decoder_blk_nums = [3, 3, 3, 3]
    self.encoder_channels = [32, 64, 128, 256]
    self.decoder_channels = [256, 128, 64, 32]

  def _octree_align(self, value: torch.Tensor, octree: OctreeD,
                    octree_query: OctreeD, depth: int):
    key = octree.graphs[depth].key
    query = octree_query.graphs[depth].key
    assert key.shape[0] == value.shape[0]
    return ocnn.nn.search_value(value, key, query)

  def octree_encoder(self, octree: OctreeD):
    convs = dict()  # graph convolution features
    depth = octree.depth
    data = octree.get_input_feature(feature=self.feature)
    convs[depth] = self.conv1(data, octree, depth)
    for i in range(self.encoder_stages):
      d = depth - i
      convs[d] = self.encoder[i](convs[d], octree, d)
      if i < self.encoder_stages - 1:
        convs[d-1] = self.downsample[i](convs[d], octree, d)
    return convs

  def octree_decoder(self, convs: dict, octree_in: OctreeD, octree_out: OctreeD,
                     update_octree: bool = False):
    logits, signals = dict(), dict()
    depth = octree_in.depth - self.encoder_stages + 1
    deconv = convs[depth]
    for i in range(self.decoder_stages):
      d = depth + i
      if i > 0:
        deconv = self.upsample[i-1](deconv, octree_out, d-1)
        skip = self._octree_align(convs[d], octree_in, octree_out, d)
        deconv = deconv + skip  # output-guided skip connections
        deconv = self.decoder[i-1](deconv, octree_out, d)

      # predict the splitting label and signal
      logit = self.predict[i](deconv, octree_out, d)
      nnum = octree_out.nnum[d]
      logits[d] = logit[-nnum:]

      # regress signals and pad zeros to non-leaf nodes
      signal = self.regress[i](deconv, octree_out, d)
      signals[d] = self.graph_pad(signal, octree_out, d)

      # update the octree according to predicted labels
      if update_octree:
        split = logits[d].argmax(1).int()
        octree_out.octree_split(split, d)
        if i < self.decoder_stages - 1:
          octree_out.octree_grow(d + 1)

    return {'logits': logits, 'signals': signals, 'octree_out': octree_out}

  def forward(self, octree_in: OctreeD, octree_out: OctreeD,
              pos: torch.Tensor = None, update_octree: bool = False):
    # run encoder and decoder
    convs = self.octree_encoder(octree_in)
    output = self.octree_decoder(convs, octree_in, octree_out, update_octree)

    # setup mpu
    depth_out = octree_out.depth
    neural_mpu = mpu.NeuralMPU(output['signals'], octree_out, depth_out)

    # compute function value with mpu
    if pos is not None:
      output['mpus'] = neural_mpu(pos)

    # create the mpu wrapper
    output['neural_mpu'] = lambda pos: neural_mpu(pos)[depth_out]
    return output



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import torch

from ognn import mpu
from ognn.octreed import OctreeD
from ognn.ounet import GraphOUNet


class GraphUNet(GraphOUNet):

  def __init__(self, in_channels: int, resblk_type: str = 'basic',
               feature: str = 'L', norm_type: str = 'batch_norm',
               act_type: str = 'relu', **kwargs):
    super().__init__(
        in_channels, resblk_type, feature, norm_type, act_type)
    self.predict = None

  def config_network(self):
    self.n_node_type = 7
    self.encoder_blk_nums = [2, 2, 2, 2, 2, 2]
    self.decoder_blk_nums = [2, 2, 2, 2, 2, 2]
    self.encoder_channels = [32, 32, 32, 64, 128, 256]
    self.decoder_channels = [256, 128, 64, 32, 32, 32]

  def octree_decoder(self, convs: dict, octree: OctreeD):
    signals = dict()
    depth = octree.depth - self.encoder_stages + 1
    deconv = convs[depth]
    for i in range(self.decoder_stages):
      d = depth + i
      if i > 0:
        deconv = self.upsample[i-1](deconv, octree, d-1)
        deconv = deconv + convs[d]  # skip connections
        deconv = self.decoder[i-1](deconv, octree, d)

      # regress signals and pad zeros to non-leaf nodes
      signal = self.regress[i](deconv, octree, d)
      signals[d] = self.graph_pad(signal, octree, d)

    return {'signals': signals}

  def forward(self, octree: OctreeD, octree_out: OctreeD,
              pos: torch.Tensor = None, update_octree: bool = False):
    # run encoder and decoder
    convs = self.octree_encoder(octree)
    output = self.octree_decoder(convs, octree)

    # setup mpu
    depth_out = octree.depth
    neural_mpu = mpu.NeuralMPU(output['signals'], octree, depth_out)

    # compute function value with mpu
    if pos is not None:
      output['mpus'] = neural_mpu(pos)

    # create the mpu wrapper
    output['neural_mpu'] = lambda pos: neural_mpu(pos)[depth_out]
    return output



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import ocnn
import torch


def collate_func(batch):
  output = ocnn.dataset.CollateBatch(merge_points=False)(batch)

  if 'pos' in output:
    batch_idx = torch.cat([torch.ones(pos.size(0), 1) * i
                           for i, pos in enumerate(output['pos'])], dim=0)
    pos = torch.cat(output['pos'], dim=0)
    output['pos'] = torch.cat([pos, batch_idx], dim=1)

  for key in ['grad', 'sdf', 'occu', 'weight', 'color']:
    if key in output:
      output[key] = torch.cat(output[key], dim=0)

  return output



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import math
import ocnn
import torch
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint

from ognn.octreed import OctreeD
from ognn.utils import scatter_mean


class Activation(torch.nn.Module):

  def __init__(self, act_type: str = 'relu', inplace: bool = True):
    super().__init__()
    self.act_type = act_type.lower()
    if self.act_type == 'relu':
      self.activation = torch.nn.ReLU(inplace)
    elif self.act_type == 'silu':
      self.activation = torch.nn.SiLU(inplace)
    elif self.act_type == 'gelu':
      self.activation = torch.nn.GELU()
    else:
      raise ValueError

  def forward(self, x: torch.Tensor):
    return self.activation(x)


class GraphConv(torch.nn.Module):

  def __init__(
          self, in_channels: int, out_channels: int, n_edge_type: int = 7,
          n_node_type: int = 0, use_bias: bool = False):
    super().__init__()
    self.avg_degree = 7
    self.in_channels = in_channels
    self.out_channels = out_channels
    self.n_edge_type = n_edge_type
    self.n_node_type = n_node_type
    self.use_bias = use_bias

    node_channel = n_node_type if n_node_type > 1 else 0
    self.weights = torch.nn.Parameter(
        torch.Tensor(n_edge_type * (in_channels + node_channel), out_channels))
    if self.use_bias:
      self.bias = torch.nn.Parameter(torch.Tensor(out_channels))
    self.reset_parameters()

  def reset_parameters(self) -> None:
    fan_in = self.avg_degree * self.in_channels
    fan_out = self.avg_degree * self.out_channels
    std = math.sqrt(2.0 / float(fan_in + fan_out))
    a = math.sqrt(3.0) * std
    torch.nn.init.uniform_(self.weights, -a, a)
    if self.use_bias:
      torch.nn.init.zeros_(self.bias)

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    graph = octree.graphs[depth]

    # concatenate the one_hot vector for node_type
    if self.n_node_type > 1:
      one_hot = F.one_hot(graph.node_type, num_classes=self.n_node_type)
      x = torch.cat([x, one_hot], dim=1)

    # x -> col_data
    row, col = graph.edge_idx
    index = row * self.n_edge_type + graph.edge_type
    col_data = scatter_mean(
        x[col], index, dim=0, dim_size=x.shape[0] * self.n_edge_type)

    # add self-loops
    index = torch.arange(graph.nnum, dtype=torch.int64, device=x.device)
    index = index * self.n_edge_type + (self.n_edge_type - 1)
    col_data[index] = x

    # matrix product
    output = col_data.view(x.shape[0], -1) @ self.weights

    # add bias
    if self.use_bias:
      out += self.bias
    return output

  def extra_repr(self) -> str:
    return ('in_channels={}, out_channels={}, n_edge_type={}, n_node_type={}, '
            'use_bias={}'.format(self.in_channels, self.out_channels,
             self.n_edge_type, self.n_node_type, self.use_bias))  # noqa


class GraphNorm(torch.nn.Module):

  def __init__(self, in_channels: int, norm_type: str = 'batch_norm'):
    super().__init__()

    self.in_channels = in_channels
    self.norm_type = norm_type

    self.group = 32
    self.min_group_channels = 4
    if self.min_group_channels * self.group > in_channels:
      self.group = in_channels // self.min_group_channels
    assert in_channels % self.group == 0

    if self.norm_type == 'batch_norm':
      self.norm = torch.nn.BatchNorm1d(in_channels)  # , bn_eps, bn_momentum)
    elif self.norm_type == 'group_norm':
      self.norm = ocnn.nn.OctreeGroupNorm(in_channels, self.group)
    else:
      raise ValueError

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    if self.norm_type == 'batch_norm':
      output = self.norm(x)
    elif self.norm_type == 'group_norm':
      output = self.norm(x, octree, depth)
    else:
      raise ValueError
    return output


class GraphConvNorm(torch.nn.Module):

  def __init__(
          self, in_channels: int, out_channels: int, n_edge_type: int = 7,
          n_node_type: int = 0, norm_type: str = 'batch_norm'):
    super().__init__()
    self.conv = GraphConv(in_channels, out_channels, n_edge_type, n_node_type)
    self.norm = GraphNorm(out_channels, norm_type)

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    out = self.conv(x, octree, depth)
    out = self.norm(out, octree, depth)
    return out


class GraphConvNormAct(torch.nn.Module):

  def __init__(
          self, in_channels: int, out_channels: int, n_edge_type: int = 7,
          n_node_type: int = 0, norm_type: str = 'batch_norm',
          act_type: str = 'relu'):
    super().__init__()
    self.conv = GraphConv(in_channels, out_channels, n_edge_type, n_node_type)
    self.norm = GraphNorm(out_channels, norm_type)
    self.act = Activation(act_type=act_type)

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    out = self.conv(x, octree, depth)
    out = self.norm(out, octree, depth)
    out = self.act(out)
    return out


class GraphPad(torch.nn.Module):

  def __init__(self):
    super().__init__()

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    # pad zeros to get features for the original octree nodes
    octree_mask = octree.graphs[depth].octree_mask
    out = x.new_zeros(octree_mask.shape[0], x.shape[1])
    out[octree_mask] = x  # pad zeros for internal octree nodes
    return out


class Conv1x1(torch.nn.Module):

  def __init__(
          self, in_channels: int, out_channels: int, use_bias: bool = False):
    super().__init__()
    self.linear = torch.nn.Linear(in_channels, out_channels, use_bias)

  def forward(self, x):
    return self.linear(x)


class Conv1x1Norm(torch.nn.Module):

  def __init__(self, in_channels: int, out_channels: int,
               norm_type: str = 'batch_norm'):
    super().__init__()
    self.conv = Conv1x1(in_channels, out_channels, use_bias=False)
    self.norm = GraphNorm(out_channels, norm_type)
    # self.bn = torch.nn.BatchNorm1d(out_channels, bn_eps, bn_momentum)

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    out = self.conv(x)
    out = self.norm(out, octree, depth)
    # out = self.bn(out) #, octree, depth)
    return out


class Conv1x1NormAct(torch.nn.Module):

  def __init__(self, in_channels: int, out_channels: int,
               norm_type: str = 'batch_norm', act_type: str = 'relu'):
    super().__init__()
    self.conv = Conv1x1(in_channels, out_channels, use_bias=False)
    self.norm = GraphNorm(out_channels, norm_type)
    self.act = Activation(act_type)

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    out = self.conv(x)
    out = self.norm(out, octree, depth)
    out = self.act(out)
    return out


class Prediction(torch.nn.Module):

  def __init__(self, in_channels: int, mid_channels: int, out_channels: int,
               norm_type: str = 'batch_norm', act_type: str = 'relu'):
    super().__init__()
    self.conv1 = Conv1x1NormAct(in_channels, mid_channels, norm_type, act_type)
    self.conv2 = Conv1x1(mid_channels, out_channels, use_bias=True)

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    x = self.conv1(x, octree, depth)
    x = self.conv2(x)
    return x


class Upsample(torch.nn.Module):

  def __init__(self, in_channels: int):
    super().__init__()
    self.in_channels = in_channels
    self.weights = torch.nn.Parameter(torch.Tensor(in_channels, in_channels, 8))
    torch.nn.init.xavier_uniform_(self.weights)
    # TODO: add bias

  def forward(self, x: torch.Tensor):
    out = x @ self.weights.flatten(1)
    out = out.view(-1, self.in_channels)
    return out

  def extra_repr(self):
    return 'in_channels={}'.format(self.in_channels)


class Downsample(torch.nn.Module):

  def __init__(self, in_channels):
    super().__init__()
    self.in_channels = in_channels

    self.weights = torch.nn.Parameter(torch.Tensor(in_channels, in_channels, 8))
    torch.nn.init.xavier_uniform_(self.weights)

  def forward(self, x: torch.Tensor):
    weights = self.weights.flatten(1).t()
    out = x.view(-1, self.in_channels * 8) @ weights
    return out

  def extra_repr(self):
    return 'in_channels={}'.format(self.in_channels)


class GraphDownsample(torch.nn.Module):

  def __init__(self, in_channels: int, out_channels: int,
               norm_type: str = 'batch_norm', act_type: str = 'relu'):
    super().__init__()
    self.in_channels = in_channels
    self.out_channels = out_channels
    self.downsample = Downsample(in_channels)
    if in_channels != out_channels:
      self.conv1x1 = Conv1x1NormAct(
          in_channels, out_channels, norm_type, act_type)

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    # downsample nodes at layer depth
    numd = octree.nnum[depth]
    outd = self.downsample(x[-numd:])

    # get the nodes at layer (depth-1)
    out = x.new_zeros(octree.nnum[depth-1], x.shape[1])
    leaf_mask = octree.children[depth-1] < 0
    leaf_num = octree.lnum[depth-1]
    out[leaf_mask] = x[-leaf_num-numd:-numd]
    out[leaf_mask.logical_not()] = outd

    # construct the final output
    out = torch.cat([x[:-leaf_num-numd], out], dim=0)

    if self.in_channels != self.out_channels:
      out = self.conv1x1(out, octree, depth - 1)
    return out

  def extra_repr(self):
    return 'in_channels={}, out_channels={}'.format(
        self.in_channels, self.out_channels)


class GraphUpsample(torch.nn.Module):

  def __init__(self, in_channels: int, out_channels: int,
               norm_type: str = 'batch_norm', act_type: str = 'relu'):
    super().__init__()
    self.in_channels = in_channels
    self.out_channels = out_channels
    self.upsample = Upsample(in_channels)
    if in_channels != out_channels:
      self.conv1x1 = Conv1x1NormAct(
          in_channels, out_channels, norm_type, act_type)

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    # upsample nodes at layer depth
    numd = octree.nnum[depth]
    leaf_mask = octree.children[depth] < 0
    outd = x[-numd:]
    out1 = self.upsample(outd[leaf_mask.logical_not()])

    # construct the final output
    out = torch.cat([x[:-numd], outd[leaf_mask], out1], dim=0)
    if self.in_channels != self.out_channels:
      out = self.conv1x1(out, octree, depth + 1)
    return out

  def extra_repr(self):
    return 'in_channels={}, out_channels={}'.format(
        self.in_channels, self.out_channels)


class GraphResBlock2(torch.nn.Module):

  def __init__(self, in_channels: int, out_channels: int, n_edge_type: int = 7,
               n_node_type: int = 0, norm_type: str = 'batch_norm',
               act_type: str = 'relu', bottleneck: int = 4):
    super().__init__()
    self.in_channels = in_channels
    self.out_channels = out_channels
    self.bottleneck = bottleneck
    channel_m = int(out_channels / bottleneck)

    self.conva = GraphConvNormAct(
        in_channels, channel_m, n_edge_type, n_node_type, norm_type, act_type)
    self.convb = GraphConvNorm(
        channel_m, out_channels, n_edge_type, n_node_type, norm_type)
    if self.in_channels != self.out_channels:
      self.conv1x1 = Conv1x1Norm(in_channels, out_channels, norm_type)
    self.act = Activation(act_type)

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    x1 = self.conva(x, octree, depth)
    x2 = self.convb(x1, octree, depth)

    if self.in_channels != self.out_channels:
      x = self.conv1x1(x, octree, depth)

    out = self.act(x2 + x)
    return out


class GraphResBlock(torch.nn.Module):

  def __init__(self, in_channels: int, out_channels: int, n_edge_type: int = 7,
               n_node_type: int = 0, norm_type: str = 'batch_norm',
               act_type: str = 'relu', bottleneck: int = 4):
    super().__init__()
    self.in_channels = in_channels
    self.out_channels = out_channels
    self.bottleneck = bottleneck
    channel_m = int(out_channels / bottleneck)

    self.conv1x1a = Conv1x1NormAct(in_channels, channel_m, norm_type, act_type)
    self.conv = GraphConvNormAct(
        channel_m, channel_m, n_edge_type, n_node_type, norm_type, act_type)
    self.conv1x1b = Conv1x1Norm(channel_m, out_channels, norm_type)
    if self.in_channels != self.out_channels:
      self.conv1x1c = Conv1x1Norm(in_channels, out_channels, norm_type)
    self.act = Activation(act_type)

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    x1 = self.conv1x1a(x, octree, depth)
    x2 = self.conv(x1, octree, depth)
    x3 = self.conv1x1b(x2, octree, depth)

    if self.in_channels != self.out_channels:
      x = self.conv1x1c(x, octree, depth)

    out = self.act(x3 + x)
    return out


class GraphResBlocks(torch.nn.Module):

  def __init__(self, in_channels: int, out_channels: int, n_edge_type: int = 7,
               n_node_type: int = 0, norm_type: str = 'batch_norm',
               act_type: str = 'relu', bottleneck: int = 4, resblk_num: int = 1,
               resblk_type: str = 'basic', use_checkpoint: bool = True):
    super().__init__()
    self.resblk_num = resblk_num
    self.use_checkpoint = use_checkpoint
    channels = [in_channels] + [out_channels] * resblk_num
    ResBlk = self._get_resblock(resblk_type)
    self.resblks = torch.nn.ModuleList([ResBlk(channels[i], channels[i+1],
        n_edge_type, n_node_type, norm_type, act_type, bottleneck)
        for i in range(self.resblk_num)])  # noqa

  def _get_resblock(self, resblk_type):
    if resblk_type == 'bottleneck':
      return GraphResBlock
    elif resblk_type == 'basic':
      return GraphResBlock2
    else:
      raise ValueError

  def forward(self, x: torch.Tensor, octree: OctreeD, depth: int):
    for i in range(self.resblk_num):
      if self.use_checkpoint and self.training:
        x = checkpoint(self.resblks[i], x, octree, depth, use_reentrant=False)
      else:
        x = self.resblks[i](x, octree, depth)
    return x



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

from ognn import nn
from ognn.ounet import GraphOUNet
from ognn.octreed import OctreeD


class GraphAE(GraphOUNet):

  def __init__(self, in_channels: int, resblk_type: str = 'basic',
               feature: str = 'L', norm_type: str = 'batch_norm',
               act_type: str = 'relu', code_channel: int = 8, **kwargs):
    super().__init__(
        in_channels, resblk_type, feature, norm_type, act_type)

    # reduce the code channel for fair comparison; if the code is of high
    # dimension, the performance will be significantly better
    self.code_channel = code_channel
    # code_dim = self.code_channel * 2 ** (3 * full_depth)

    self.project1 = nn.Conv1x1NormAct(
        self.encoder_channels[-1], self.code_channel,
        self.norm_type, self.act_type)
    self.project2 = nn.Conv1x1NormAct(
        self.code_channel, self.decoder_channels[0],
        self.norm_type, self.act_type)

  def octree_encoder(self, octree: OctreeD):
    convs = super().octree_encoder(octree)
    depth = octree.depth - self.encoder_stages + 1

    # reduce the dimension to get the shape code
    shape_code = self.project1(convs[depth], octree, depth)
    return shape_code

  def octree_decoder(self, shape_code, octree_in: OctreeD, octree_out: OctreeD,
                     update_octree: bool = False):
    logits, signals = dict(), dict()
    depth = octree_in.depth - self.encoder_stages + 1
    deconv = self.project2(shape_code, octree_out, depth)
    for i in range(self.decoder_stages):
      d = depth + i
      if i > 0:
        deconv = self.upsample[i-1](deconv, octree_out, d-1)
        # there is no skip connections between the encoder and decoder
        deconv = self.decoder[i-1](deconv, octree_out, d)

      # predict the splitting label and signal
      logit = self.predict[i](deconv, octree_out, d)
      nnum = octree_out.nnum[d]
      logits[d] = logit[-nnum:]

      # regress signals and pad zeros to non-leaf nodes
      signal = self.regress[i](deconv, octree_out, d)
      signals[d] = self.graph_pad(signal, octree_out, d)

      # update the octree according to predicted labels
      if update_octree:
        split = logits[d].argmax(1).int()
        octree_out.octree_split(split, d)
        if i < self.decoder_stages - 1:
          octree_out.octree_grow(d + 1)

    return {'logits': logits, 'signals': signals, 'octree_out': octree_out}



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import os
import torch
import numpy as np

from thsolver import Dataset
from ocnn.octree import Octree, Points
from .utils import collate_func


class TransformShape:

  def __init__(self, flags):
    self.flags = flags

    self.volume_sample_num = flags.volume_sample_num
    self.surface_sample_num = flags.surface_sample_num
    self.points_scale = 0.5  # the points are in [-0.5, 0.5]

    self.depth = flags.depth
    self.full_depth = flags.full_depth

  def points2octree(self, points: Points):
    octree = Octree(self.depth, self.full_depth)
    octree.build_octree(points)
    return octree

  def process_points_cloud(self, sample):
    # get the input
    points = torch.from_numpy(sample['points']).float()
    normals = torch.from_numpy(sample['normals']).float()
    colors = torch.from_numpy(sample['colors']).float()
    points = points / self.points_scale  # scale to [-1.0, 1.0]

    # transform points to octree
    features = colors if self.flags.load_color else None
    points_gt = Points(points=points, normals=normals, features=features)
    points_gt.clip(min=-1, max=1)
    octree_gt = self.points2octree(points_gt)

    # construct the output dict
    return {'octree_in': octree_gt, 'octree_gt': octree_gt, 'points_in': points}

  def sample_volume(self, sample):
    sdf = sample['sdf']
    grad = sample['grad']
    points = sample['points'] / self.points_scale  # to [-1, 1]

    rand_idx = np.random.choice(points.shape[0], size=self.volume_sample_num)
    points = torch.from_numpy(points[rand_idx]).float()
    sdf = torch.from_numpy(sdf[rand_idx]).float()
    grad = torch.from_numpy(grad[rand_idx]).float()
    color = torch.zeros(self.volume_sample_num, 3)
    return {'pos': points, 'sdf': sdf, 'grad': grad, 'color': color}

  def sample_surface(self, sample):
    normals = sample['normals']
    colors = sample['colors']
    points = sample['points'] / self.points_scale  # to [-1, 1]

    rand_idx = np.random.choice(points.shape[0], size=self.surface_sample_num)
    points = torch.from_numpy(points[rand_idx]).float()
    colors = torch.from_numpy(colors[rand_idx]).float()
    normals = torch.from_numpy(normals[rand_idx]).float()
    sdf = torch.zeros(self.surface_sample_num)
    return {'pos': points, 'sdf': sdf, 'grad': normals, 'color': colors}

  def __call__(self, sample, idx):
    output = self.process_points_cloud(sample['point_cloud'])

    samples = self.sample_volume(sample['sdf'])
    surface = self.sample_surface(sample['point_cloud'])
    for key in samples.keys():
      samples[key] = torch.cat([samples[key], surface[key]], dim=0)
    if not self.flags.load_color:
      samples.pop('color', None)

    output.update(samples)
    return output


class ReadFile:

  def __init__(self, load_color=False):
    self.load_color = load_color

  def __call__(self, filename):
    filename_pc = os.path.join(filename, 'pointcloud.npz')
    raw = np.load(filename_pc)
    point_cloud = {'points': raw['points'], 'normals': raw['normals']}

    filename_sdf = os.path.join(filename, 'sdf.npz')
    raw = np.load(filename_sdf)
    sdf = {'points': raw['points'], 'grad': raw['grad'], 'sdf': raw['sdf']}

    if self.load_color:
      filename_color = os.path.join(filename, 'color.npz')
      raw = np.load(filename_color)
      point_cloud['colors'] = raw['colors']  # colors of the input point clouds
    else:
      point_cloud['colors'] = np.zeros_like(point_cloud['normals'])

    return {'point_cloud': point_cloud, 'sdf': sdf}


def get_shapenet_vae_dataset(flags):
  transform = TransformShape(flags)
  read_file = ReadFile(flags.load_color)
  dataset = Dataset(flags.location, flags.filelist, transform,
                    read_file=read_file, in_memory=flags.in_memory)
  return dataset, collate_func



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import torch
import numpy as np
from plyfile import PlyData

from thsolver import Dataset
from ocnn.octree import Octree, Points
from .utils import collate_func


class Transform:
  r'''Load point clouds from ply files, rescale the points and build octree.
  Used to evaluate the network trained on ShapeNet.'''

  def __init__(self, flags):
    self.flags = flags
    self.depth = flags.depth
    self.full_depth = flags.full_depth
    self.point_scale = flags.point_scale

  def points2octree(self, points: Points):
    octree = Octree(self.depth, self.full_depth)
    octree.build_octree(points)
    return octree

  def __call__(self, points, idx):
    # After normalization, the points are in [-1, 1]
    points = points[:, :3] / self.point_scale

    # construct the points and octree
    points_in = Points(points=torch.from_numpy(points).float(),
                       features=torch.ones(points.shape[0], 1).float())
    points_in.clip(-1.0, 1.0)
    octree_in = self.points2octree(points_in)

    return {'points_in': points, 'octree_in': octree_in}


def read_file(filename: str):
  plydata = PlyData.read(filename + '.ply')
  vtx = plydata['vertex']
  points = np.stack([vtx['x'], vtx['y'], vtx['z']], axis=1).astype(np.float32)
  return points


def get_pointcloud_eval_dataset(flags):
  transform = Transform(flags)
  dataset = Dataset(flags.location, flags.filelist, transform,
                    read_file=read_file, in_memory=flags.in_memory)
  return dataset, collate_func



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import torch
from typing import Optional
from ocnn.utils import broadcast, scatter_add


def scatter_mean(src: torch.Tensor, index: torch.Tensor, dim: int = -1,
                 out: Optional[torch.Tensor] = None,
                 dim_size: Optional[int] = None) -> torch.Tensor:
  out = scatter_add(src, index, dim, out, dim_size)
  dim_size = out.size(dim)

  index_dim = dim
  if index_dim < 0:
    index_dim = index_dim + src.dim()
  if index.dim() <= index_dim:
    index_dim = index.dim() - 1

  ones = torch.ones(index.size(), dtype=src.dtype, device=src.device)
  count = scatter_add(ones, index, index_dim, None, dim_size)
  count[count < 1] = 1
  count = broadcast(count, out, dim)
  if out.is_floating_point():
    out.true_divide_(count)
  else:
    out.div_(count, rounding_mode='floor')
  return out


def spmm(index, value, m, n, matrix):
  r'''Matrix product of sparse matrix with dense matrix.

  Args:
      index (:class:`LongTensor`): The index tensor of sparse matrix.
      value (:class:`Tensor`): The value tensor of sparse matrix.
      m (int): The first dimension of corresponding dense matrix.
      n (int): The second dimension of corresponding dense matrix.
      matrix (:class:`Tensor`): The dense matrix.

  :rtype: :class:`Tensor`
  '''

  assert n == matrix.size(-2)

  row, col = index
  matrix = matrix if matrix.dim() > 1 else matrix.unsqueeze(-1)

  out = matrix.index_select(-2, col)
  out = out * value.unsqueeze(-1)
  out = scatter_add(out, row, dim=-2, dim_size=m)

  return out



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import ocnn
import torch
import torch.nn.functional as F


def compute_gradient(y, x):
  grad_outputs = torch.ones_like(y)
  grad = torch.autograd.grad(y, [x], grad_outputs, create_graph=True)[0]
  return grad


def sdf_reg_loss(sdf, grad, sdf_gt, grad_gt, name_suffix=''):
  wg, ws = 1.0, 200.0
  grad_loss = (grad - grad_gt).pow(2).mean() * wg
  sdf_loss = (sdf - sdf_gt).pow(2).mean() * ws
  loss_dict = {'grad_loss' + name_suffix: grad_loss,
               'sdf_loss' + name_suffix: sdf_loss}
  return loss_dict


def sdf_grad_loss(sdf, grad, sdf_gt, grad_gt, name_suffix=''):
  on_surf = sdf_gt != -1
  off_surf = on_surf.logical_not()

  sdf_loss = sdf[on_surf].pow(2).mean() * 200.0
  norm_loss = (grad[on_surf] - grad_gt[on_surf]).pow(2).mean() * 1.0
  intr_loss = torch.exp(-40 * torch.abs(sdf[off_surf])).mean() * 0.1
  grad_loss = (grad[off_surf].norm(2, dim=-1) - 1).abs().mean() * 0.1

  losses = [sdf_loss, intr_loss, norm_loss, grad_loss]
  names = ['sdf_loss', 'inter_loss', 'norm_loss', 'grad_loss']
  names = [name + name_suffix for name in names]
  loss_dict = dict(zip(names, losses))
  return loss_dict


def sdf_grad_regularized_loss(sdf, grad, sdf_gt, grad_gt, name_suffix=''):
  on_surf = sdf_gt != -1
  off_surf = on_surf.logical_not()

  sdf_loss = sdf[on_surf].pow(2).mean() * 200.0
  norm_loss = (grad[on_surf] - grad_gt[on_surf]).pow(2).mean() * 1.0
  intr_loss = torch.exp(-40 * torch.abs(sdf[off_surf])).mean() * 0.1
  grad_loss = (grad[off_surf].norm(2, dim=-1) - 1).abs().mean() * 0.1
  grad_reg_loss = (grad[off_surf] - grad_gt[off_surf]).pow(2).mean() * 0.1

  losses = [sdf_loss, intr_loss, norm_loss, grad_loss, grad_reg_loss]
  names = ['sdf_loss', 'inter_loss', 'norm_loss', 'grad_loss', 'grad_reg_loss']
  names = [name + name_suffix for name in names]
  loss_dict = dict(zip(names, losses))
  return loss_dict


def possion_grad_loss(sdf, grad, sdf_gt, grad_gt, name_suffix=''):
  on_surf = sdf_gt == 0
  out_of_bbox = sdf_gt == 1.0
  off_surf = on_surf.logical_not()

  sdf_loss = sdf[on_surf].pow(2).mean() * 200.0
  norm_loss = (grad[on_surf] - grad_gt[on_surf]).pow(2).mean() * 1.0
  intr_loss = torch.exp(-40 * torch.abs(sdf[off_surf])).mean() * 0.1
  grad_loss = grad[off_surf].pow(2).mean() * 0.1  # poisson loss
  bbox_loss = torch.mean(torch.relu(-sdf[out_of_bbox])) * 100.0

  losses = [sdf_loss, intr_loss, norm_loss, grad_loss, bbox_loss]
  names = ['sdf_loss', 'inter_loss', 'norm_loss', 'grad_loss', 'bbox_loss']
  names = [name + name_suffix for name in names]
  loss_dict = dict(zip(names, losses))
  return loss_dict


def compute_mpu_gradients(mpus, pos, fval_transform=None):
  grads = dict()
  for d in mpus.keys():
    fval = mpus[d]
    if fval_transform is not None:
      fval = fval_transform(fval)
    grads[d] = compute_gradient(fval, pos)[:, :3]
  return grads


def compute_octree_loss(logits, octree_out: ocnn.octree.Octree):
  weights = [1.0] * 16
  # weights = [1.0] * 4 + [0.8, 0.6, 0.4] + [0.2] * 16

  output = dict()
  for d in logits.keys():
    logitd = logits[d]
    label_gt = octree_out.nempty_mask(d).long()
    output['loss_%d' % d] = F.cross_entropy(logitd, label_gt) * weights[d]
    output['accu_%d' % d] = logitd.argmax(1).eq(label_gt).float().mean()
  return output


def compute_sdf_loss(mpus, grads, sdf_gt, grad_gt, reg_loss_func):
  output = dict()
  for d in mpus.keys():
    sdf = mpus[d]  # TODO: tune the loss weights and `flgs`
    reg_loss = reg_loss_func(sdf, grads[d], sdf_gt, grad_gt, '_%d' % d)
    # if d < 3:  # ignore depth 2
    #   for key in reg_loss.keys():
    #     reg_loss[key] = reg_loss[key] * 0.0
    output.update(reg_loss)
  return output


def compute_occu_loss_1214(mpus, occu):
  # tried on 2021.12.13
  weights = [0.2] * 4 + [0.4, 0.6, 0.8] + [1.0] * 16  # TODO: tune the weights

  inside = occu == 0
  outside = occu == 1
  output = dict()
  for d in mpus.keys():
    sdf = mpus[d]

    inside_loss = torch.mean(torch.relu(sdf[inside])) * (1000 * weights[d])
    outside_loss = torch.mean(torch.relu(-sdf[outside])) * (1000 * weights[d])

    output['inside_loss_%d' % d] = inside_loss
    output['outside_loss_%d' % d] = outside_loss
    output['inside_accu_%d' % d] = (sdf[inside] < 0).float().mean()
    output['outside_accu_%d' % d] = (sdf[outside] > 0).float().mean()
  return output


def compute_occu_loss(mpus, grads, occu, grad_gt):
  weights = [1.0] * 16
  # weights = [0.2] * 4 + [0.4, 0.6, 0.8] + [1.0] * 16
  # weights = [0.0] * 7 + [1.0] * 16  # Single level loss

  inside = occu == 0
  outside = occu == 1
  on_surf = occu == 0.5
  off_surf = on_surf.logical_not()

  output = dict()
  for d in mpus.keys():
    sdf = mpus[d]
    grad = grads[d]
    grad_diff = grad[on_surf] - grad_gt[on_surf]

    sdf_loss = sdf[on_surf].pow(2).mean() * (200 * weights[d])
    norm_loss = grad_diff.pow(2).mean() * (1.0 * weights[d])
    intr_loss = torch.exp(-40 * sdf[off_surf].abs()).mean() * (0.1 * weights[d])
    grad_loss = grad[off_surf].pow(2).mean() * (0.1 * weights[d])

    inside_loss = torch.mean(torch.relu(sdf[inside])) * (500 * weights[d])
    outside_loss = torch.mean(torch.relu(-sdf[outside])) * (2000 * weights[d])
    inside_accu = (sdf[inside] < 0).float().mean()
    outside_accu = (sdf[outside] > 0).float().mean()

    losses = [sdf_loss, norm_loss, grad_loss, intr_loss,
              inside_loss, inside_accu, outside_loss, outside_accu]
    names = ['sdf_loss', 'norm_loss', 'grad_loss', 'inter_loss',
             'inside_loss', 'inside_accu', 'outside_loss', 'outside_accu']
    names = [name + ('_%d' % d) for name in names]
    loss_dict = dict(zip(names, losses))
    output.update(loss_dict)

  return output


def compute_occu_loss_cls(mpus, grads, occu_gt, grad_gt):
  weights = [1.0] * 16
  # weights = [0.2] * 4 + [0.4, 0.6, 0.8] + [1.0] * 16
  # weights = [0.0] * 7 + [1.0] * 16  # Single level loss

  inside = occu_gt == 0
  outside = occu_gt == 1
  on_surf = occu_gt == 0.5
  off_surf = on_surf.logical_not()

  # Use soft-version occupancies
  occu_soft = torch.ones_like(occu_gt) * 0.5
  occu_soft[inside] = 0.3
  occu_soft[outside] = 0.7

  output = dict()
  for d in mpus.keys():
    sdf = mpus[d]
    grad = grads[d]
    grad_diff = grad[on_surf] - grad_gt[on_surf]

    sdf_loss = sdf[on_surf].pow(2).mean() * (200 * weights[d])
    norm_loss = grad_diff.pow(2).mean() * (1.0 * weights[d])
    intr_loss = torch.exp(-40 * sdf[off_surf].abs()).mean() * (0.1 * weights[d])
    grad_loss = grad[off_surf].pow(2).mean() * (0.1 * weights[d])

    loss_o = F.binary_cross_entropy_with_logits(sdf, occu_soft)
    accu = (sdf.sigmoid() > 0.5).eq(occu_gt).float()
    inside_accu = accu[inside].mean()
    outside_accu = accu[outside].mean()

    losses = [sdf_loss, norm_loss, grad_loss, intr_loss,
              loss_o, inside_accu, outside_accu]
    names = ['sdf_loss', 'norm_loss', 'grad_loss', 'inter_loss',
             'occu_loss', 'inside_accu', 'outside_accu']
    names = [name + ('_%d' % d) for name in names]
    loss_dict = dict(zip(names, losses))
    output.update(loss_dict)

  return output


def compute_color_loss(colors, color_gt, on_surf=None):
  output = dict()
  for d in colors.keys():
    color_loss = (colors[d] - color_gt)[on_surf].pow(2).mean() * 20.0
    output['color_loss_%d' % d] = color_loss
  return output


def get_sdf_loss_function(loss_type=''):
  if loss_type == 'sdf_reg_loss':
    return sdf_reg_loss
  elif loss_type == 'sdf_grad_loss':
    return sdf_grad_loss
  elif loss_type == 'possion_grad_loss':
    return possion_grad_loss
  elif loss_type == 'sdf_grad_reg_loss':
    return sdf_grad_regularized_loss
  else:
    return None


def shapenet_loss(batch, model_out, reg_loss_type='', **kwargs):
  # octree loss
  output = compute_octree_loss(model_out['logits'], model_out['octree_out'])

  # regression loss
  grads = compute_mpu_gradients(model_out['mpus'], batch['pos'])
  reg_loss_func = get_sdf_loss_function(reg_loss_type)
  sdf_loss = compute_sdf_loss(
      model_out['mpus'], grads, batch['sdf'], batch['grad'], reg_loss_func)
  output.update(sdf_loss)
  return output


def dfaust_loss(batch, model_out, reg_loss_type='', **kwargs):
  # there is no octree loss
  grads = compute_mpu_gradients(model_out['mpus'], batch['pos'])
  reg_loss_func = get_sdf_loss_function(reg_loss_type)
  output = compute_sdf_loss(
      model_out['mpus'], grads, batch['sdf'], batch['grad'], reg_loss_func)
  return output


def synthetic_room_loss(batch, model_out, *args, **kwargs):
  # octree loss
  output = compute_octree_loss(model_out['logits'], model_out['octree_out'])

  # grads
  grads = compute_mpu_gradients(model_out['mpus'], batch['pos'])

  # occu loss
  occu_loss = compute_occu_loss(
      model_out['mpus'], grads, batch['occu'], batch['grad'])
  output.update(occu_loss)

  return output


def shapenet_vae_loss(batch, model_out, reg_loss_type='', **kwargs):
  output = shapenet_loss(batch, model_out, reg_loss_type)

  if 'colors' in model_out:
    color_loss = compute_color_loss(model_out['colors'], batch['color'])
    output.update(color_loss)

  if 'kl_loss' in model_out.keys():
    kl_weight = kwargs['kl_weight']
    output['kl_loss'] = kl_weight * model_out['kl_loss']
  return output


def shapenet_vae_color_loss(batch, model_out, reg_loss_type='', **kwargs):
  mpus = model_out['mpus']
  sdfs, colors = dict(), dict()
  for k in mpus.keys():
    sdfs[k] = mpus[k][:, 0]
    colors[k] = mpus[k][:, 1:]
  model_out['mpus'] = sdfs  # to reuse the original loss functions
  model_out['colors'] = colors

  output = shapenet_loss(batch, model_out, reg_loss_type)

  on_surf = batch['sdf'] == 0
  color_loss = compute_color_loss(model_out['colors'], batch['color'], on_surf)
  output.update(color_loss)

  if 'kl_loss' in model_out.keys():
    kl_weight = kwargs['kl_weight']
    output['kl_loss'] = kl_weight * model_out['kl_loss']
  return output



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import torch
from typing import List

from ognn import nn
from ognn import mpu
from ognn.octreed import OctreeD


class TinyEncoder(torch.nn.Module):

  def __init__(self, channels: List[int], resblk_nums: List[int],
               resblk_type: str = 'basic', bottleneck: int = 2,
               n_node_type: int = -1, **kwargs):
    super().__init__()
    n_edge_type = 7
    act_type = 'gelu'
    norm_type = 'group_norm'
    use_checkpoint = True
    self.stage_num = len(channels)
    self.delta_depth = len(channels) - 1

    n_node_types = [n_node_type - i for i in range(self.stage_num)]
    self.encoder = torch.nn.ModuleList([nn.GraphResBlocks(
        channels[i], channels[i], n_edge_type, n_node_types[i], norm_type,
        act_type, bottleneck, resblk_nums[i], resblk_type, use_checkpoint)
        for i in range(self.stage_num)])
    self.downsample = torch.nn.ModuleList([nn.GraphDownsample(
        channels[i], channels[i+1], norm_type, act_type)
        for i in range(self.stage_num - 1)])  # Note: self.stage_num - 1

  def forward(self, data: torch.Tensor, octree: OctreeD, depth: int):
    out = dict()
    out[depth] = data
    for i in range(self.stage_num):
      d = depth - i
      out[d] = self.encoder[i](out[d], octree, d)
      if i < self.stage_num - 1:
        out[d-1] = self.downsample[i](out[d], octree, d)
    return out


class TinyDecoder(torch.nn.Module):

  def __init__(self, channels: List[int], resblk_nums: List[int],
               resblk_type: str = 'basic', bottleneck: int = 2,
               n_node_type: int = -1, **kwargs):
    super().__init__()
    n_edge_type = 7
    act_type = 'gelu'
    norm_type = 'group_norm'
    use_checkpoint = True
    self.stage_num = len(channels)

    n_node_types = [n_node_type + i for i in range(self.stage_num)]
    self.decoder = torch.nn.ModuleList([nn.GraphResBlocks(
        channels[i], channels[i], n_edge_type, n_node_types[i], norm_type,
        act_type, bottleneck, resblk_nums[i], resblk_type, use_checkpoint)
        for i in range(self.stage_num)])
    self.upsample = torch.nn.ModuleList([nn.GraphUpsample(
        channels[i], channels[i+1], norm_type, act_type)
        for i in range(self.stage_num - 1)])

  def forward(self, datas: torch.Tensor, octree: OctreeD, depth: int):
    out = datas[depth]
    for i in range(self.stage_num):
      d = depth + i
      out = self.decoder[i](out, octree, d)
      if i < self.stage_num - 1:
        out = self.upsample[i](out, octree, d)
        out = out + datas[d+1]  # skip connections
    return out


class OctreeDecoder(torch.nn.Module):

  def __init__(self, channels: List[int], resblk_nums: List[int],
               resblk_type: str = 'basic', bottleneck: int = 2,
               n_node_type: int = -1, out_channels: int = 4,
               predict_octree: bool = False, **kwargs):
    super().__init__()
    n_edge_type = 7
    act_type = 'gelu'
    norm_type = 'group_norm'
    use_checkpoint = True
    mid_channels = 32
    self.stage_num = len(channels)
    self.predict_octree = predict_octree

    n_node_types = [n_node_type + i for i in range(self.stage_num)]
    self.decoder = torch.nn.ModuleList([nn.GraphResBlocks(
        channels[i], channels[i], n_edge_type, n_node_types[i], norm_type,
        act_type, bottleneck, resblk_nums[i], resblk_type, use_checkpoint)
        for i in range(self.stage_num)])
    self.upsample = torch.nn.ModuleList([nn.GraphUpsample(
        channels[i], channels[i+1], norm_type, act_type)
        for i in range(self.stage_num - 1)])
    self.graph_pad = nn.GraphPad()

    self.regress = torch.nn.ModuleList([nn.Prediction(
        channels[i], mid_channels, out_channels, norm_type, act_type)
        for i in range(self.stage_num)])
    if predict_octree:
      self.predict = torch.nn.ModuleList([nn.Prediction(
          channels[i], mid_channels, 2, norm_type, act_type)
          for i in range(self.stage_num)])

  def forward(self, data: torch.Tensor, octree: OctreeD, depth: int,
              update_octree: bool = False):
    logits, signals = dict(), dict()
    for i in range(self.stage_num):
      d = depth + i
      data = self.decoder[i](data, octree, d)

      # predict the splitting label and signal
      if self.predict_octree:
        logit = self.predict[i](data, octree, d)
        nnum = octree.nnum[d]
        logits[d] = logit[-nnum:]

      # regress signals and pad zeros to non-leaf nodes
      signal = self.regress[i](data, octree, d)
      signals[d] = self.graph_pad(signal, octree, d)

      # update the octree according to predicted labels
      if update_octree and self.predict_octree:
        split = logits[d].argmax(1).int()
        octree.octree_split(split, d)
        if i < self.stage_num - 1:
          octree.octree_grow(d + 1)

      # upsample
      if i < self.stage_num - 1:
        data = self.upsample[i](data, octree, d)

    return {'logits': logits, 'signals': signals, 'octree_out': octree}


class TinyNet(torch.nn.Module):

  def __init__(self, channels: List[int], resblk_nums: List[int],
               resblk_type: str = 'basic', bottleneck: int = 2,
               n_node_type: int = -1, **kwargs):
    super().__init__()
    self.encoder = TinyEncoder(
        channels, resblk_nums, resblk_type, bottleneck, n_node_type)
    self.decoder = TinyDecoder(
        channels[::-1], resblk_nums[::-1], resblk_type, bottleneck,
        n_node_type - self.encoder.delta_depth)

  def forward(self, data: torch.Tensor, octree: OctreeD, depth: int):
    encs = self.encoder(data, octree, depth)
    out = self.decoder(encs, octree, depth - self.encoder.delta_depth)
    return out


class Encoder(torch.nn.Module):

  def __init__(self, in_channels: int, n_node_type: int = 7,
               enc_channels: List[int] = [24, 32, 64],
               enc_resblk_nums: List[int] = [1, 1, 1],
               net_channels: List[int] = [64, 128, 256],
               net_resblk_nums: List[int] = [1, 2, 2], **kwargs):
    super().__init__()
    self.delta_depth = len(enc_channels) - 1

    n_edge_type = 7
    act_type = 'gelu'
    norm_type = 'group_norm'
    resblk_type = 'basic'

    self.conv1 = nn.GraphConvNormAct(
        in_channels, enc_channels[0], n_edge_type, n_node_type,
        norm_type, act_type)
    self.encoding = TinyEncoder(
        enc_channels, enc_resblk_nums, resblk_type, bottleneck=4,
        n_node_type=n_node_type)
    self.net = TinyNet(
        net_channels, net_resblk_nums, resblk_type, bottleneck=2,
        n_node_type=n_node_type - self.delta_depth)

  def forward(self, data: torch.Tensor, octree: OctreeD, depth: int):
    conv = self.conv1(data, octree, depth)
    convs = self.encoding(conv, octree, depth)
    depth = depth - self.delta_depth
    data = self.net(convs[depth], octree, depth)
    return data


class Decoder(torch.nn.Module):

  def __init__(self, out_channels: int, n_node_type: int = 7,
               dec_channels: List[int] = [64, 32, 24],
               dec_resblk_nums: List[int] = [1, 1, 1],
               net_channels: List[int] = [64, 128, 256],
               net_resblk_nums: List[int] = [1, 2, 2],
               predict_octree: bool = False, **kwargs):
    super().__init__()

    self.delta_depth = len(dec_channels) - 1
    n_node_type = n_node_type - self.delta_depth
    self.net = TinyNet(
        net_channels, net_resblk_nums, resblk_type='basic',
        bottleneck=2, n_node_type=n_node_type)
    self.decoding = OctreeDecoder(
        dec_channels, dec_resblk_nums, resblk_type='basic',
        bottleneck=4, n_node_type=n_node_type, out_channels=out_channels,
        predict_octree=predict_octree)

  def forward(self, data: torch.Tensor, octree: OctreeD, depth: int,
              update_octree: bool = False):
    data = self.net(data, octree, depth)
    output = self.decoding(data, octree, depth, update_octree)
    return output


class DiagonalGaussianDistribution(torch.nn.Module):

  def __init__(self, parameters: torch.Tensor, deterministic: bool = False):
    super().__init__()
    self.deterministic = deterministic
    self.parameters = parameters

    self.device = parameters.device
    self.mean, self.logvar = torch.chunk(parameters, 2, dim=1)
    self.logvar = torch.clamp(self.logvar, -30.0, 20.0)
    self.std = torch.exp(0.5 * self.logvar)
    self.var = torch.exp(self.logvar)
    if self.deterministic:
      self.var = torch.zeros_like(self.mean).to(device=self.device)
      self.std = self.var.clone()

  def sample(self):
    x = self.mean + self.std * torch.randn(self.mean.shape, device=self.device)
    return x

  def kl(self, other=None):
    if self.deterministic:
      out = torch.Tensor([0.])
    else:
      if other is None:
        out = 0.5 * (torch.pow(self.mean, 2) + self.var - 1.0 - self.logvar)
      else:
        out = 0.5 * torch.sum(
            torch.pow(self.mean - other.mean, 2) / other.var +
            self.var / other.var - 1.0 - self.logvar + other.logvar,
            dim=[1, 2, 3, 4])
    return out

  def nll(self, sample, dims=[1, 2, 3, 4]):
    if self.deterministic:
      out = torch.Tensor([0.])
    else:
      logtwopi = 1.8378770664093453  # np.log(2.0 * np.pi)
      out = 0.5 * torch.sum(
          logtwopi + self.logvar + torch.pow(sample - self.mean, 2) / self.var,
          dim=dims)
    return out

  def mode(self):
    return self.mean


class GraphVAE(torch.nn.Module):

  def __init__(self, in_channels: int, n_node_type: int = 7,
               code_channels: int = 3, out_channels: int = 4,
               feature: str = 'ND', **kwargs):
    super().__init__()
    self.feature = feature
    self.config_network()

    self.encoder = Encoder(
        in_channels, n_node_type, self.enc_channels, self.enc_resblk_nums,
        self.enc_net_channels, self.enc_net_resblk_nums)
    self.decoder = Decoder(
        out_channels, n_node_type, self.dec_channels, self.dec_resblk_nums,
        self.dec_net_channels, self.dec_net_resblk_nums, predict_octree=True)

    self.pre_kl_conv = nn.Conv1x1(
        self.enc_channels[-1], 2 * code_channels, use_bias=True)
    self.post_kl_conv = nn.Conv1x1(
        code_channels, self.dec_channels[0], use_bias=True)

  def config_network(self):
    self.enc_channels = [24, 32, 32]
    self.enc_resblk_nums = [1, 1, 1]
    self.enc_net_channels = [32, 64, 256]
    self.enc_net_resblk_nums = [1, 1, 1]

    self.dec_channels = [32, 32, 24]
    self.dec_resblk_nums = [1, 1, 1]
    self.dec_net_channels = [32, 64, 256]
    self.dec_net_resblk_nums = [1, 1, 1]

  def forward(self, octree_in: OctreeD, octree_out: OctreeD,
              pos: torch.Tensor = None, update_octree: bool = False):
    code = self.extract_code(octree_in)
    posterior = DiagonalGaussianDistribution(code)
    z = posterior.sample()
    code_depth = octree_in.depth - self.encoder.delta_depth
    # The input paramter `octree_out` is not used. It is just for compatibility
    # with the other models.
    output = self.decode_code(z, code_depth, octree_in, pos, update_octree)

    output['kl_loss'] = posterior.kl().mean()
    output['code_max'] = z.max()
    output['code_min'] = z.min()
    return output

  def extract_code(self, octree_in: OctreeD):
    depth = octree_in.depth
    data = octree_in.get_input_feature(feature=self.feature)
    conv = self.encoder(data, octree_in, depth)
    code = self.pre_kl_conv(conv)    # project features to the vae code
    return code

  def decode_code(self, code: torch.Tensor, code_depth: int, octree: OctreeD,
                  pos: torch.Tensor = None, update_octree: bool = False):
    data = self.post_kl_conv(code)   # project the vae code to features
    output = self.decoder(data, octree, code_depth, update_octree)

    # setup mpu
    depth_out = octree.depth
    neural_mpu = mpu.NeuralMPU(output['signals'], octree, depth_out)

    # compute function value with mpu
    if pos is not None:
      output['mpus'] = neural_mpu(pos)

    # create the mpu wrapper
    output['neural_mpu'] = lambda pos: neural_mpu(pos)[depth_out]
    return output



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

from . import nn
from . import loss

from .octreed import OctreeD
from .ounet import GraphOUNet
from .ae import GraphAE
from .unet import GraphUNet
from .vae import GraphVAE
from .mpu import NeuralMPU

__all__ = [
    'nn', 'loss', 'OctreeD',
    'GraphOUNet', 'GraphAE', 'GraphUNet', 'GraphVAE', 'NeuralMPU',
]



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

from .shapenet import get_shapenet_dataset
from .pointcloud import get_pointcloud_dataset, get_singlepointcloud_dataset
from .pointcloud_eval import get_pointcloud_eval_dataset
from .synthetic_room import get_synthetic_room_dataset
from .shapenet_vae import get_shapenet_vae_dataset


__all__ = [
    'get_shapenet_dataset',
    'get_pointcloud_dataset',
    'get_singlepointcloud_dataset',
    'get_pointcloud_eval_dataset',
    'get_synthetic_room_dataset',
    'get_shapenet_vae_dataset'
]



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import torch
import numpy as np
from ocnn.octree import Octree, Points
from thsolver import Dataset

from .utils import collate_func


class Transform:

  def __init__(self, flags):
    self.flags = flags
    self.depth = flags.depth
    self.full_depth = flags.full_depth
    self.point_sample_num = flags.point_sample_num
    self.point_scale = flags.point_scale

  def build_octree(self, points):
    pts, normals = points[:, :3], points[:, 3:]
    points_cloud = Points(pts, normals)
    octree = Octree(self.depth, self.full_depth)
    octree.build_octree(points_cloud)
    return {'octree_in': octree, 'points_in': points, 'octree_gt': octree}

  def sample_on_surface(self, points):
    '''Randomly sample points on the surface.'''

    rnd_idx = torch.randint(high=points.shape[0], size=(self.point_sample_num,))
    pos = points[rnd_idx, :3]
    normal = points[rnd_idx, 3:]
    sdf = torch.zeros(self.point_sample_num)
    return {'pos': pos, 'grad': normal, 'sdf': sdf}

  def sample_off_surface(self, bbox):
    '''Randomly sample points in the 3D space.'''

    # uniformly sampling in the whole 3D sapce
    pos = torch.rand(self.point_sample_num, 3) * 2 - 1

    # point gradients
    # grad = torch.zeros(self.point_sample_num, 3)
    norm = torch.sqrt(torch.sum(pos**2, dim=1, keepdim=True)) + 1e-6
    grad = pos / norm  # !!! fake off-surface gradients

    # sdf values
    esp = 0.04
    bbmin, bbmax = bbox[:3] - esp, bbox[3:] + esp
    mask = torch.logical_and(pos > bbmin, pos < bbmax).all(1)  # inbox
    sdf = -1.0 * torch.ones(self.point_sample_num)  # !!! fake off-surface sdf
    sdf[mask.logical_not()] = 1.0  # exactly out-of-bbox
    return {'pos': pos, 'grad': grad, 'sdf': sdf}

  def scale_and_clip_points(self, points):
    points[:, :3] = points[:, :3] * self.point_scale  # rescale points
    pts = points[:, :3]
    mask = torch.logical_and(pts > -1.0, pts < 1.0).all(1)
    return points[mask]

  def compute_bbox(self, points):
    pts = points[:, :3]
    bbmin = pts.min(0)[0] - 0.06
    bbmax = pts.max(0)[0] + 0.06
    bbmin = torch.clamp(bbmin, min=-1, max=1)
    bbmax = torch.clamp(bbmax, min=-1, max=1)
    return torch.cat([bbmin, bbmax])

  def __call__(self, points, idx):
    points = self.scale_and_clip_points(points)  # clip points to [-1, 1]
    output = self.build_octree(points)
    bbox = self.compute_bbox(points)
    output['bbox'] = bbox  # used in marching cubes

    sdf_on_surf = self.sample_on_surface(points)
    sdf_off_surf = self.sample_off_surface(bbox)
    sdfs = {key: torch.cat([sdf_on_surf[key], sdf_off_surf[key]], dim=0)
            for key in sdf_on_surf.keys()}
    output.update(sdfs)
    return output


class SingleTransform(Transform):

  def __init__(self, flags):
    super().__init__(flags)
    self.output = None
    self.points = None
    self.bbox = None

  def __call__(self, points, idx):
    if self.output is None:
      self.points = self.scale_and_clip_points(points)  # clip points to [-1, 1]
      self.output = self.build_octree(self.points)
      self.bbox = self.compute_bbox(self.points)
      self.output['bbox'] = self.bbox  # used in marching cubes

    output = self.output
    sdf_on_surf = self.sample_on_surface(self.points)
    sdf_off_surf = self.sample_off_surface(self.bbox)
    sdfs = {key: torch.cat([sdf_on_surf[key], sdf_off_surf[key]], dim=0)
            for key in sdf_on_surf.keys()}
    output.update(sdfs)
    return output


def read_file(filename: str):
  if filename.endswith('.xyz'):
    points = np.loadtxt(filename)
  elif filename.endswith('.npy'):
    points = np.load(filename)
  else:
    raise NotImplementedError
  output = torch.from_numpy(points.astype(np.float32))
  return output


def get_pointcloud_dataset(flags):
  transform = Transform(flags)
  dataset = Dataset(flags.location, flags.filelist, transform,
                    read_file=read_file, in_memory=flags.in_memory)
  return dataset, collate_func


def get_singlepointcloud_dataset(flags):
  transform = SingleTransform(flags)
  dataset = Dataset(flags.location, flags.filelist, transform,
                    read_file=read_file, in_memory=flags.in_memory)
  return dataset, collate_func



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

import os
import torch
import numpy as np

from thsolver import Dataset
from ocnn.octree import Octree, Points
from .utils import collate_func


class TransformShape:

  def __init__(self, flags):
    self.flags = flags

    self.point_sample_num = 3000
    self.sdf_sample_num = 5000
    self.points_scale = 0.5  # the points are in [-0.5, 0.5]
    self.noise_std = 0.005

    self.depth = flags.depth
    self.full_depth = flags.full_depth

  def points2octree(self, points: Points):
    octree = Octree(self.depth, self.full_depth)
    octree.build_octree(points)
    return octree

  def process_points_cloud(self, sample):
    # get the input
    points, normals = sample['points'], sample['normals']
    points = points / self.points_scale  # scale to [-1.0, 1.0]

    # transform points to octree
    points_gt = Points(points=torch.from_numpy(points).float(),
                       normals=torch.from_numpy(normals).float())
    points_gt.clip(min=-1, max=1)
    octree_gt = self.points2octree(points_gt)

    if self.flags.distort:
      # randomly sample points and add noise
      # Since we rescale points to [-1.0, 1.0] in Line 24, we also need to
      # rescale the `noise_std` here to make sure the `noise_std` is always
      # 0.5% of the bounding box size.
      noise_std = self.noise_std / self.points_scale
      noise = noise_std * np.random.randn(self.point_sample_num, 3)
      rand_idx = np.random.choice(points.shape[0], size=self.point_sample_num)
      points_noise = points[rand_idx] + noise

      points_in = Points(points=torch.from_numpy(points_noise).float(),
                         features=torch.ones(self.point_sample_num, 1).float())
      points_in.clip(-1.0, 1.0)
      octree_in = self.points2octree(points_in)
    else:
      points_in = points_gt
      octree_in = octree_gt

    # construct the output dict
    return {'octree_in': octree_in, 'points_in': points_in.points,
            'octree_gt': octree_gt, 'points_gt': points_gt}

  def sample_sdf(self, sample):
    sdf = sample['sdf']
    grad = sample['grad']
    points = sample['points'] / self.points_scale  # to [-1, 1]

    rand_idx = np.random.choice(points.shape[0], size=self.sdf_sample_num)
    points = torch.from_numpy(points[rand_idx]).float()
    sdf = torch.from_numpy(sdf[rand_idx]).float()
    grad = torch.from_numpy(grad[rand_idx]).float()
    return {'pos': points, 'sdf': sdf, 'grad': grad}

  def sample_on_surface(self, points, normals):
    rand_idx = np.random.choice(points.shape[0], size=self.sdf_sample_num)
    xyz = torch.from_numpy(points[rand_idx]).float()
    grad = torch.from_numpy(normals[rand_idx]).float()
    sdf = torch.zeros(self.sdf_sample_num)
    return {'pos': xyz, 'sdf': sdf, 'grad': grad}

  def sample_off_surface(self, xyz):
    xyz = xyz / self.points_scale  # to [-1, 1]

    rand_idx = np.random.choice(xyz.shape[0], size=self.sdf_sample_num)
    xyz = torch.from_numpy(xyz[rand_idx]).float()
    # grad = torch.zeros(self.sample_number, 3)  # dummy grads
    grad = xyz / (xyz.norm(p=2, dim=1, keepdim=True) + 1.0e-6)
    sdf = -1 * torch.ones(self.sdf_sample_num)  # dummy sdfs
    return {'pos': xyz, 'sdf': sdf, 'grad': grad}

  def __call__(self, sample, idx):
    output = self.process_points_cloud(sample['point_cloud'])

    # sample ground truth sdfs
    if self.flags.load_sdf:
      sdf_samples = self.sample_sdf(sample['sdf'])
      output.update(sdf_samples)

    # sample on surface points and off surface points
    if self.flags.sample_surf_points:
      on_surf = self.sample_on_surface(sample['points'], sample['normals'])
      off_surf = self.sample_off_surface(sample['sdf']['points'])  # TODO
      sdf_samples = {
          'pos': torch.cat([on_surf['pos'], off_surf['pos']], dim=0),
          'grad': torch.cat([on_surf['grad'], off_surf['grad']], dim=0),
          'sdf': torch.cat([on_surf['sdf'], off_surf['sdf']], dim=0)}
      output.update(sdf_samples)

    return output


class ReadFile:
  def __init__(self, load_sdf=False, load_occu=False):
    self.load_occu = load_occu
    self.load_sdf = load_sdf

  def __call__(self, filename):
    filename_pc = os.path.join(filename, 'pointcloud.npz')
    raw = np.load(filename_pc)
    point_cloud = {'points': raw['points'], 'normals': raw['normals']}
    output = {'point_cloud': point_cloud}

    if self.load_occu:
      filename_occu = os.path.join(filename, 'points.npz')
      raw = np.load(filename_occu)
      occu = {'points': raw['points'], 'occupancies': raw['occupancies']}
      output['occu'] = occu

    if self.load_sdf:
      filename_sdf = os.path.join(filename, 'sdf.npz')
      raw = np.load(filename_sdf)
      sdf = {'points': raw['points'], 'grad': raw['grad'], 'sdf': raw['sdf']}
      output['sdf'] = sdf
    return output


def get_shapenet_dataset(flags):
  transform = TransformShape(flags)
  read_file = ReadFile(flags.load_sdf, flags.load_occu)
  dataset = Dataset(flags.location, flags.filelist, transform,
                    read_file=read_file, in_memory=flags.in_memory)
  return dataset, collate_func



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------

from setuptools import setup, find_packages

__version__ = '1.1.1'

with open("README.md", "r", encoding="utf-8") as fid:
  long_description = fid.read()

setup(
    name='ognn',
    version=__version__,
    author='Peng-Shuai Wang',
    author_email='wangps@hotmail.com',
    url='https://github.com/octree-nn/ognn-pytorch',
    description='Dual Octree Graph Neural Networks',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=['torch', 'numpy', 'ocnn', 'easydict'],
    python_requires='>=3.6',
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



# --------------------------------------------------------
# Dual Octree Graph Neural Networks
# Copyright (c) 2024 Peng-Shuai Wang <wangps@hotmail.com>
# Licensed under The MIT License [see LICENSE for details]
# Written by Peng-Shuai Wang
# --------------------------------------------------------


import ocnn
import torch
import torch.nn

from ocnn.octree import Octree, xyz2key
from ocnn.utils import scatter_add, range_grid
from easydict import EasyDict as edict


class NeuralMPU(object):

  def __init__(self, features: torch.Tensor, octree: Octree, depth_end: int):
    super().__init__()
    self.kNN = 8
    self.buffer = (features, octree, depth_end)  # used for MPU computation
    feature = features[depth_end]
    assert feature.dim() == 2 and feature.size(1) % 4 == 0
    self.channel_out = feature.size(1) // 4

  def linear_basis(self, x: torch.Tensor):
    return 1.0 - x.abs()

  def perpare(self, octree: Octree, depth: int, pts: torch.Tensor):
    r''' Neighborhood searching and weight computation for MPU. '''

    # compute the coordinates of neighboring octree nodes
    scale = 2 ** depth
    mask = range_grid(0, 1, pts.device)
    xyzf, ids = torch.split(pts, [3, 1], 1)
    xyzf = (xyzf + 1.0) * (scale / 2.0)    # [-1, 1] -> [0, scale]
    xyzf = xyzf - 0.5                      # the code is defined on the center
    xyzi = torch.floor(xyzf).detach()      # the integer part (N, 3), use floor
    corners = xyzi.unsqueeze(1) + mask     # (N, 8, 3)
    coordsf = xyzf.unsqueeze(1) - corners  # (N, 8, 3), in [-1.0, 1.0]

    # coorers -> key -> the indices of neighboring octree nodes
    xyz = corners.view(-1, 3)
    ids = ids.detach().repeat(1, self.kNN).view(-1)  # (N, 8, 1) -> (N*8, )
    key = xyz2key(xyz[:, 0], xyz[:, 1], xyz[:, 2], ids, depth)
    idx = octree.search_key(key, depth)

    # corners -> in-bound flags
    inbound = torch.logical_and(corners > -1, corners < scale)
    inbound = torch.all(inbound, dim=-1).view(-1)
    valid = torch.logical_and(idx > -1, inbound)  # in-bound and idx > -1

    # the indices of the pts itself
    ids = torch.arange(pts.size(0), device=pts.device, dtype=torch.long)
    ids = ids.unsqueeze(-1).repeat(1, self.kNN).view(-1)
    ids = ids[valid]

    # remove invalid pts
    idx = idx[valid].long()               # (N*8, )   -> (N', )
    coordsf = coordsf.view(-1, 3)[valid]  # (N, 8, 3) -> (N', 3)

    # bspline weights
    weights = self.linear_basis(coordsf)                # (N', 3)
    weights = torch.prod(weights, axis=-1).view(-1)     # (N', )

    # Here, the scale factor `2**(depth - 6)` is used to emphasize high-resolution
    # basis functions. Tune this factor further if needed! !!! NOTE !!!
    # weights = weights * 2**(depth - 6)                 # used for shapenet
    weights = weights * (depth**2 / 50)                  # testing

    # rescale back the original scale.
    # after recaling, the coordsf is in the same scale as `pts``
    coordsf = coordsf * (2.0 / scale)   # [-1.0, 1.0] -> [-2.0/scale, 2.0/scale]
    return edict({'ids': ids, 'idx': idx, 'xyzf': coordsf, 'weights': weights})

  def compute(self, pts: torch.Tensor, feature: torch.Tensor, octree: Octree,
              mpus: edict, depth_start: int, depth_end: int):
    npt = pts.size(0)
    indices, weights, xyzfs = [], [], []
    nnum_cum = ocnn.utils.cumsum(octree.nnum, dim=0, exclusive=True)
    for d in range(depth_start, depth_end+1):
      idxd = mpus[d].idx
      idsd = mpus[d].ids
      xyzfd = mpus[d].xyzf
      weightd = mpus[d].weights

      if d < depth_end:
        child = octree.children[d]
        leaf = child[idxd] < 0  # keep only leaf nodes
        idsd = idsd[leaf]
        idxd = idxd[leaf]
        weightd = weightd[leaf]
        xyzfd = xyzfd[leaf]

      idxd = idxd + (nnum_cum[d] - nnum_cum[depth_start])
      indices.append(torch.stack([idsd, idxd], dim=1))
      weights.append(weightd)
      xyzfs.append(xyzfd)

    indices = torch.cat(indices, dim=0).t()
    weights = torch.cat(weights, dim=0)
    xyzfs = torch.cat(xyzfs, dim=0)
    output = self.spmm(indices, weights, npt, feature.size(0), feature, xyzfs)
    return output

  def spmm(self, index, weights, m, n, feature, xyzf):
    row, col = index
    assert n == feature.size(0)
    feature = feature if feature.dim() > 1 else feature.unsqueeze(-1)

    k = xyzf.shape[0]  # number of non-zero elements of the sparse matrix
    ones = torch.ones((k, 1), device=xyzf.device)
    xyz1 = torch.cat([xyzf, ones], dim=1).unsqueeze(1)   # (K, 1, 4)
    out = feature.index_select(0, col).view(k, -1, 4)    # (K, C, 4)
    out = torch.sum(out * xyz1, dim=2, keepdim=False)    # (K, C)
    out = out * weights.unsqueeze(1)                     # (K, C) * (K, 1)
    out = scatter_add(out, row, dim=0, dim_size=m)       # (M, C)

    norm = ones * weights.unsqueeze(1)                   # (K, 1) * (K, 1)
    norm = scatter_add(norm, row, dim=0, dim_size=m)     # (M, 1)

    out = torch.div(out, norm + 1e-8).squeeze()          # (M, C)
    return out

  def __call__(self, pts: torch.Tensor):
    features, octree, depth_end = self.buffer

    fvals, mpus = dict(), dict()
    depth_start = octree.full_depth
    for d in range(depth_start, depth_end+1):
      mpus[d] = self.perpare(octree, d, pts)
      if d in features:
        fvals[d] = self.compute(pts, features[d], octree, mpus, depth_start, d)
    return fvals
